# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_ml_helpers.ipynb.

# %% ../nbs/02_ml_helpers.ipynb 4
from __future__ import annotations
from .utils import *
from .chart_plotting import plot_permutation_importances,plot_confusion_matrix,plot_residuals,plot_feature_importances
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV, cross_validate, train_test_split
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier,AdaBoostClassifier
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor,AdaBoostRegressor
from sklearn.metrics import f1_score,accuracy_score,classification_report,log_loss
import numpy as np
import pandas as pd
import statsmodels.api as sm
import re

# %% auto 0
__all__ = ['run_logistic_regression', 'run_multinomial_statmodel', 'run_sklearn_model', 'tune_sklearn_model', 'do_param_search',
           'summarize_cv_results', 'summarize_default_cv', 'show_both_cv', 'get_adaboost_info']

# %% ../nbs/02_ml_helpers.ipynb 5
def run_logistic_regression(X_trn:pd.DataFrame, # Training dataframe
                            y_trn:pd.Series|np.ndarray, # Training label
                            multi_class='multinomial', # sklearn's log reg multiclass option
                            solver='newton-cg', # sklearn's log reg solver option
                            penalty=None, # sklearn's log reg penalty option
                            max_iter=10000 # sklearn's log reg max iteration option
                           ):
    "Perform Sklearn logistic regression, then print coefficients and classification report"
    model = LogisticRegression(random_state=0, multi_class=multi_class, 
                               penalty=penalty, solver=solver,max_iter=max_iter).fit(X_trn, y_trn)
    preds = model.predict(X_trn)
    prob_preds = model.predict_proba(X_trn)
    print('-'*100)
    print('Intercept: \n', model.intercept_)
    print('Coefficients: \n', model.coef_)
    print('Coefficients exp :\n',np.exp(model.coef_))

    print('-'*100)
    print('Log loss: ',log_loss(y_trn,prob_preds))
    print('-'*100)
    print(classification_report(y_trn,preds))
    return model

# %% ../nbs/02_ml_helpers.ipynb 7
def run_multinomial_statmodel(X_trn:pd.DataFrame, # Training dataframe
                              y_trn:pd.Series|np.ndarray, # Training label
                              add_constant=True # To add a constant column to X_trn
                             ):
    "Perform multinominal logit from statsmodel, then print results and classification report"
    if add_constant:
        X_trn = sm.add_constant(X_trn)
    logit_model=sm.MNLogit(y_trn,X_trn)
    result=logit_model.fit()
    stats1=result.summary()
    print(stats1)
    prob_preds = logit_model.predict(params = result.params.values)
    print('-'*100)
    print('Log loss: ',log_loss(y_trn,prob_preds))
    print('-'*100)
    print(classification_report(y_trn,np.argmax(prob_preds,axis=1)))
    return logit_model

# %% ../nbs/02_ml_helpers.ipynb 9
def run_sklearn_model(model_name:str, # sklearn's Machine Learning model to try. Currently support DecisionTree,AdaBoost,RandomForest
                      model_params:dict, # A dictionary containing model's hyperparameters
                      X_trn:pd.DataFrame, # Training dataframe
                      y_trn:pd.Series|np.ndarray, # Training label
                      is_regression=False, # To use regression model or classification model
                      class_names:list=None, # List of names associated with the labels (same order); e.g. ['no','yes']. For classification only
                      test_split=None, # Test set split. If float: random split. If list of list: indices of train and test set. If None: skip splitting
                      metric_funcs={}, # Dictionary of metric functions: {metric_name:metric_func}
                      seed=42, # Random seed
                      plot_fea_imp=True # To whether plot sklearn's feature importances. Set to False to skip the plot
                    ):
    np.random.seed(seed)
    if isinstance(test_split,float):
        X_trn,X_test,y_trn,y_test = train_test_split(X_trn,y_trn,test_size=test_split,random_state=seed)
    if isinstance(test_split,list) and len(test_split)==2 and isinstance(test_split[0],list) and isinstance(test_split[1],list):
        X_test = X_trn.iloc[test_split[1]].copy()
        y_test = y_trn[test_split[1]]
        y_trn = y_trn[test_split[0]]
        X_trn = X_trn.iloc[test_split[0]]

    if model_name=='DecisionTree':
        if is_regression:
            _model = DecisionTreeRegressor(random_state=seed,**model_params)
        else:
            _model = DecisionTreeClassifier(random_state=seed,**model_params)
    elif model_name=='AdaBoost':
        dt_params={k.split('__')[1]:v for k,v in model_params.items() if 'base_estimator' in k}
        abc_params={k:v for k,v in model_params.items() if 'base_estimator' not in k}        
        print(f'Decision Tree params: {dt_params}')
        print(f'AdaBoost params: {abc_params}')
        if is_regression:
            dt = DecisionTreeRegressor(random_state=seed,**dt_params)
            _model = AdaBoostRegressor(base_estimator=dt,random_state=seed,algorithm='SAMME',**abc_params)
        else:
            dt = DecisionTreeClassifier(random_state=seed,**dt_params)
            _model = AdaBoostClassifier(base_estimator=dt,random_state=seed,**abc_params)
    elif model_name=='RandomForest':
        if is_regression:
            _model = RandomForestRegressor(random_state=seed,**model_params)
        else:
            _model = RandomForestClassifier(random_state=seed,**model_params)
    else:
        print('Unsupported model')
        return

    _model = _model.fit(X_trn,y_trn)
    pred_trn = _model.predict(X_trn)
    
    # For regression:
    if is_regression:
        print('-'*30 + ' Train set ' + '-'*30)
        for k,v in metric_funcs.items():
            print(f'{k}: {v(y_trn,pred_trn)}')
        
        if test_split is not None:
            print('-'*30 + ' Test set ' + '-'*30)
            pred_test = _model.predict(X_test)
            for k,v in metric_funcs.items():
                print(f'{k}: {v(y_test,pred_test)}')
            
            # plot residual plot
            plot_residuals(_model,X_trn,y_trn,X_test,y_test,
                           is_fit=True,
                           qqplot=True)
        else:
            plot_residuals(_model,X_trn,y_trn,
                           is_fit=True,
                           qqplot=True)

    # For classification
    else:
        prob_trn = _model.predict_proba(X_trn)
        print('-'*30 + ' Train set ' + '-'*30)
        print(f'Log loss: {log_loss(y_trn,prob_trn)}')
#         for k,v in metric_funcs.items():
#             print(f'{k}: {v(y_trn,prob_trn)}')
        print(classification_report(y_trn, pred_trn, target_names=class_names))
        if test_split is not None:
            print('-'*30 + ' Test set ' + '-'*30)
            pred_test = _model.predict(X_test)
            prob_test = _model.predict_proba(X_test)
            print(f'Log loss: {log_loss(y_test,prob_test)}')
            print(classification_report(y_test, pred_test, target_names=class_names))
            print('-'*100)
            df2 = pd.DataFrame({'Class': class_names,
                                'True Distribution':pd.Series(y_test).value_counts(normalize=True).sort_index(),
                               'Prediction Distribution':pd.Series(pred_test).value_counts(normalize=True).sort_index()}
                              )
            print(df2)
            plot_confusion_matrix(y_test,pred_test,class_names)


#     _ = plot_permutation_importances(_model,X_trn,y_trn,scoring=['neg_root_mean_squared_error'])
    if plot_fea_imp:
        plot_feature_importances(_model.feature_importances_,X_trn.columns.values)
    
    return _model

# %% ../nbs/02_ml_helpers.ipynb 11
def tune_sklearn_model(model_name:str, # sklearn's Machine Learning model to try. Currently support DecisionTree,AdaBoost,RandomForest,
                      param_grid:dict, # Dictionary with parameters names (str) as keys and lists of parameter settings to try as values
                      X_trn:pd.DataFrame, # Training dataframe
                      y_trn:pd.Series|np.ndarray, # Training label
                      is_regression=False, # Is it a regression problem, or classification?
                      custom_cv=5, # sklearn's cross-validation splitting strategy
                      random_cv_iter=None, # Number of parameter settings that are sampled. Use this if you want to do RandomizedSearchCV
                      scoring=None, # Metric
                      seed=42, # Random seed
                      rank_show=10, # Number of ranks to show (descending order)
                      show_split_scores=True, # To show both train and test split scores
                     ):
    "Perform either Sklearn's Grid Search or Randomized Search (based on random_cv_iter) of the model using param_grid"
    if is_regression:
        if model_name=='DecisionTree':
            _model = DecisionTreeRegressor(random_state=seed)
        elif model_name=='AdaBoost':
            dt = DecisionTreeRegressor(random_state=seed)
            _model = AdaBoostRegressor(base_estimator= dt,random_state=seed)
        elif model_name=='RandomForest':
            _model = RandomForestRegressor(random_state=seed)
        else:
            print('Unsupported model')
            return
    else:
        if model_name=='DecisionTree':
            _model = DecisionTreeClassifier(random_state=seed)
        elif model_name=='AdaBoost':
            dt = DecisionTreeClassifier(random_state=seed)
            _model = AdaBoostClassifier(base_estimator= dt,random_state=seed,algorithm='SAMME')
        elif model_name=='RandomForest':
            _model = RandomForestClassifier(random_state=seed)
        else:
            print('Unsupported model')
            return
    
    scoring = val2list(scoring)
    search_cv,default_cv = do_param_search(X_trn,y_trn,_model,param_grid,cv=custom_cv,scoring=scoring,random_cv_iter = random_cv_iter,seed=seed)
    # Default to show results for the first metric
    show_both_cv(search_cv,default_cv,scoring[0],rank_show,show_split_scores)
    return search_cv

# %% ../nbs/02_ml_helpers.ipynb 13
def do_param_search(
    X_train,y_train,
    estimator,
    param_grid,
    random_cv_iter=None,
    include_default=True,
    cv=None,
    scoring=None,
    seed=42
    
):
    scoring = val2list(scoring)
    search_cv,default_cv=None,None
    if random_cv_iter:
        search_cv = RandomizedSearchCV(estimator=estimator,
                                      n_iter=random_cv_iter,
                                      param_distributions=param_grid,
                                      scoring=scoring,
                                      n_jobs=-1,
                                      cv=cv,
                                        return_train_score=True,
                                      verbose=1,refit=False,random_state=seed)
        search_cv.fit(X_train,y_train)
    else:
        search_cv = GridSearchCV(estimator,param_grid,scoring=scoring,n_jobs=-1,cv=cv,verbose=1,
                                 return_train_score=True,refit=False)
        search_cv.fit(X_train,y_train)
    if include_default:
        default_cv = cross_validate(estimator,X_train,y_train,scoring=scoring,cv=cv,n_jobs=-1,verbose=1,
                                   return_train_score=True)
    return search_cv.cv_results_,default_cv
        

def summarize_cv_results(search_cv,scoring,top_n=10,show_split_scores=False):
    num_split = len([c for c in search_cv.keys() if re.search(r'split\d_test', c)])
    search_cv = pd.DataFrame(search_cv)
    search_cv = search_cv.sort_values(f'rank_test_{scoring}')
    for rec in search_cv[['params',f'mean_train_{scoring}',f'std_train_{scoring}',
                          f'mean_test_{scoring}',f'std_test_{scoring}',f'rank_test_{scoring}']+
                         [f'split{i}_train_{scoring}' for i in range(num_split)] +
                         [f'split{i}_test_{scoring}' for i in range(num_split)]
                        ].values[:top_n]:
        print('-'*10)
        print(f'Rank {rec[5]}')
        print(f'Params: {rec[0]}')
        if show_split_scores:
            print(f'Train scores: {[round(i,2) for i in rec[6:6+num_split]]}')
        print(f'Mean train score: {rec[1]:.3f} +- {rec[2]:.3f}')
        if show_split_scores:
            print(f'Test scores:  {[round(i,2) for i in rec[-num_split:]]}')
        print(f'Mean test score: {rec[3]:.3f} +- {rec[4]:.3f}')

def summarize_default_cv(default_cv,s):
    print('-'*10)
    print("Default Params")
    print(f"Mean train score: {round(default_cv[f'train_{s}'].mean(),3)} +- {round(default_cv[f'train_{s}'].std(),3)}")
    print(f"Mean test score: {round(default_cv[f'test_{s}'].mean(),3)} +- {round(default_cv[f'test_{s}'].std(),3)}")

def show_both_cv(search_cv,default_cv,scoring,top_n=10,show_split_scores=False):
    summarize_cv_results(search_cv,scoring,top_n,show_split_scores)
    summarize_default_cv(default_cv,scoring)

    
def get_adaboost_info(dt_params,ada_params,X,y,seed=42):
    dt = DecisionTreeClassifier(random_state=seed,**dt_params)
    abc = AdaBoostClassifier(base_estimator=dt,random_state=seed,**ada_params)
    abc.fit(X,y)
    for i,t in enumerate(abc.estimators_):
        print(f'{t}\n\tTree depth: {t.tree_.max_depth}, Weight: {abc.estimator_weights_[i]}, Error: {abc.estimator_errors_[i]}')
    return abc



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_chart_plotting.ipynb.

# %% ../nbs/03_chart_plotting.ipynb 3
from __future__ import annotations
from .utils import *
from pathlib import Path
import numpy as np
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
from sklearn.model_selection import learning_curve,validation_curve, RandomizedSearchCV, GridSearchCV
import matplotlib.pyplot as plt
from sklearn.inspection import PartialDependenceDisplay, permutation_importance
import pandas as pd
import plotly.express as px
import dtreeviz
from sklearn.tree import export_graphviz
import graphviz
from matplotlib.cm import get_cmap
import plotly.graph_objects as go
from statsmodels.stats.outliers_influence import variance_inflation_factor
from statsmodels.tools.tools import add_constant
import seaborn as sns
import matplotlib
from scipy.stats import chi2_contingency
from yellowbrick.regressor import ResidualsPlot

# %% auto 0
__all__ = ['get_vif', 'get_correlation_by_threshold', 'get_cat_correlation', 'plot_cat_correlation', 'plot_residuals',
           'plot_prediction_distribution', 'plot_learning_curve', 'plot_validation_curve', 'plot_tree_dtreeviz',
           'plot_classification_tree_sklearn', 'plot_feature_importances', 'plot_permutation_importances',
           'params_2D_heatmap', 'params_3D_heatmap', 'pdp_numerical_only', 'pdp_categorical_only', 'plot_ice_pair',
           'plot_confusion_matrix', 'draw_sankey']

# %% ../nbs/03_chart_plotting.ipynb 6
def get_vif(df:pd.DataFrame, # dataframe to plot
            plot_corr=False, # to plot the correlation matrix
            figsize=(10,10) # Matplotlib figsize
           ):
    """
    Perform variance inflation factor calculation, and optionally plot correlation matrix
    
    Note that your dataframe should only have numerical features to perform VIF
    """
    if plot_corr:
        fig,ax = plt.subplots(figsize=(10,10))
        dataplot = sns.heatmap(df.corr(), fmt=f".2f",cmap="YlGnBu", annot=True)  
        plt.show()

    df_const = add_constant(df)    
    vif = pd.Series([variance_inflation_factor(df_const.values, i) 
                    for i in range(df_const.shape[1])], index=df_const.columns)
    return vif

# %% ../nbs/03_chart_plotting.ipynb 11
def get_correlation_by_threshold(df_corr, # Correlation DataFrame
                         min_thres=0.98 # minimum correlation to take
                        ):
    result_dic={}
    for i,c in enumerate(df_corr.columns.tolist()):
        result = df_corr.iloc[:i,i][np.abs(df_corr.iloc[:i,i])>=min_thres].to_dict()
        result = {k:v for (k,v) in result.items() }
        if len(result)==0: continue
#         print(f'{c}: {result}')
        result_dic[c]=result
    return result_dic

# %% ../nbs/03_chart_plotting.ipynb 15
def _cramer_v(var1, var2):
    """
    Cramer’s V measures association between two nominal variables.
    Cramer’s V lies between 0 and 1 (inclusive).
    0 indicates that the two variables are not linked by any relation.
    1 indicates that there exists a strong association between the two variables.
    """
    crosstab = np.array(pd.crosstab(var1, var2))
    chi2 = chi2_contingency(crosstab)[0]
    observation, min_dim = np.sum(crosstab), min(crosstab.shape) - 1
    return np.sqrt((chi2 / observation) / min_dim)

# %% ../nbs/03_chart_plotting.ipynb 16
def get_cat_correlation(df_cat, # DataFrame with categorical features that have been processed
                       ):
    cramer_matrix = []
    cat_cols = df_cat.columns.tolist()
    for c in cat_cols:
        _tmp = []
        for c1 in cat_cols:
            _tmp.append(_cramer_v(df_cat[c],df_cat[c1]))
        cramer_matrix.append(_tmp)
    cramer_df = pd.DataFrame(cramer_matrix,columns=df_cat.columns.values)
    cramer_df.index = df_cat.columns.values
    return cramer_df

# %% ../nbs/03_chart_plotting.ipynb 17
def plot_cat_correlation(df_cat, # DataFrame with categorical features that have been processed
                        figsize=(10,10), # Matplotlib figsize
                        ):
    cramer_df = get_cat_correlation(df_cat)
    fig,ax = plt.subplots(figsize=figsize)
    dataplot = sns.heatmap(cramer_df, fmt=f".3f",cmap="YlGnBu", annot=True)  
    plt.show()

# %% ../nbs/03_chart_plotting.ipynb 28
def plot_residuals(model, # Regression model
                   X_trn, # Training dataframe
                   y_trn, # Training label
                   X_test=None, # Testing dataframe
                   y_test=None, # Testing label
                   qqplot=True, # To whether plot the qqplot
                  ):
    
    # plot residual plot
    visualizer = ResidualsPlot(model, hist=False, qqplot=qqplot)
    visualizer.fit(X_trn, y_trn)
    if not X_test is None:
        visualizer.score(X_test, y_test)
    visualizer.show()

# %% ../nbs/03_chart_plotting.ipynb 37
def plot_prediction_distribution(y_true, # True label numpy array
                                 y_pred, # Prediction numpy array
                                 figsize=(15,5) # Matplotlib figsize
                                ):
    fig,axes = plt.subplots(1,2,figsize=figsize)    
    df_plot1=pd.DataFrame()
    df_plot1['value'] = y_pred
    df_plot1['type'] = 'Predictions'
    df_plot2=pd.DataFrame()
    df_plot2['value'] = y_true
    df_plot2['type'] = 'True values'
    df_plot = pd.concat([df_plot1,df_plot2],axis=0)

    print(f'MSE: {np.mean((y_true-y_pred)**2)}')
    print(f'RMSE: {np.sqrt(np.mean((y_true-y_pred)**2))}')
    print(f"MAE: {np.mean(np.abs(y_true-y_pred))}")

    sns.histplot(df_plot, x="value", hue="type", element="step",ax=axes[0])
    sns.kdeplot(df_plot, x="value", hue="type", ax=axes[1])
    plt.show()

# %% ../nbs/03_chart_plotting.ipynb 43
def plot_learning_curve(
    estimator, # sklearn's classifier
    title, # Title of the chart
    X, # Training features
    y, # Training label
    axes=None, # matplotlib's axes
    ylim=None, # y axis range limit
    cv=None, # sklearn's cross-validation splitting strategy
    n_jobs=-1, # Number of jobs to run in parallel
    scoring=None, # metric
    train_sizes=[0.05, 0.24, 0.43, 0.62, 0.81, 1.], # List of training size portion
    save_fig=False, # To store the chart as png in images directory
    figsize=(20,5), # Matplotlib figsize
    seed=42 # Random seed
):
    if axes is None:
        _, axes = plt.subplots(1, 2, figsize=figsize)

    axes[0].set_title(title)
    if ylim is not None:
        axes[0].set_ylim(*ylim)
    axes[0].set_xlabel("Training examples")
    axes[0].set_ylabel("Score")

    train_sizes, train_scores, test_scores, fit_times, _ = learning_curve(
        estimator,
        X,
        y,
        scoring=scoring,
        cv=cv,
        n_jobs=n_jobs,
        train_sizes=train_sizes,
        return_times=True,
        random_state=seed
    )
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    fit_times_mean = np.mean(fit_times, axis=1)
    fit_times_std = np.std(fit_times, axis=1)

    # Plot learning curve
    axes[0].grid()
    axes[0].fill_between(
        train_sizes,
        train_scores_mean - train_scores_std,
        train_scores_mean + train_scores_std,
        alpha=0.1,
        color="r",
    )
    axes[0].fill_between(
        train_sizes,
        test_scores_mean - test_scores_std,
        test_scores_mean + test_scores_std,
        alpha=0.1,
        color="g",
    )
    axes[0].plot(
        train_sizes, train_scores_mean, "o-", color="r", label="Training score"
    )
    axes[0].plot(
        train_sizes, test_scores_mean, "o-", color="g", label="Cross-validation score"
    )
    axes[0].legend(loc="best")

    # Plot n_samples vs fit_times
    axes[1].grid()
    axes[1].plot(train_sizes, fit_times_mean, "o-")
    axes[1].fill_between(
        train_sizes,
        fit_times_mean - fit_times_std,
        fit_times_mean + fit_times_std,
        alpha=0.1,
    )
    axes[1].set_xlabel("Training examples")
    axes[1].set_ylabel("fit_times")
    axes[1].set_title("Scalability of the model")

    plt.grid()
    if save_fig:
        _path = Path('./images')
        create_dir(_path)
        plt.savefig(_path/f'{title}.png')
    return plt

# %% ../nbs/03_chart_plotting.ipynb 46
def plot_validation_curve(
    estimator, # sklearn's classifier
    title, # Title of the chart
    X, # Training features
    y, # Training label
    ylim=None, # y axis range limit
    cv=None, # sklearn's cross-validation splitting strategy
    param_name=None, # Name of model's hyperparameter
    param_range=None, # List containing range of value for param_name
    is_log=False, # To log the value in param_range, for plotting
    n_jobs=-1, # Number of jobs to run in parallel
    scoring=None, # metric
    save_fig=False, # To store the chart as png in images directory    
    figsize=(8,4), # Matplotlib figsize
    fill_between=True, # To add a upper and lower one-std line for train and test curve
    enumerate_x=False # Convert categorical hyperparam to numerical, for x axis
):
    train_scores, test_scores = validation_curve(
            estimator,
            X,
            y,
            param_name=param_name,
            param_range=param_range,
            scoring=scoring,
            n_jobs=n_jobs,
            cv=cv,
            )
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    plt.figure(figsize=figsize)
    plt.title(title)
    plt.xlabel(param_name)
    plt.ylabel("Score")
    if is_log:
        plt.semilogx(param_range, train_scores_mean, marker="o",label="Training score", color="r")
        plt.semilogx(param_range, test_scores_mean, marker="o",label="Cross-validation score", color="g")

    else:
        tmp = param_range
        if enumerate_x:
            tmp = np.arange(1,len(tmp)+1)
        plt.plot(tmp, train_scores_mean, "o-", color="r", label="Training score")
        plt.plot(tmp, test_scores_mean, "o-", color="g", label="Cross-validation score")

    if fill_between:
        plt.fill_between(
            param_range,
            train_scores_mean - train_scores_std,
            train_scores_mean + train_scores_std,
            alpha=0.1,
            color="r",
        )
        plt.fill_between(
            param_range,
            test_scores_mean - test_scores_std,
            test_scores_mean + test_scores_std,
            alpha=0.1,
            color="g",
        )
    if ylim is not None:
        plt.ylim(*ylim)
    plt.grid()
    plt.legend(loc="best")
    if save_fig:
        _path = Path('./images')
        create_dir(_path)
        plt.savefig(_path/f'{title}.png')
    return plt

# %% ../nbs/03_chart_plotting.ipynb 51
def plot_tree_dtreeviz(estimator, # sklearn's classifier
                       X, # Training features
                       y, # Training label
                       target_name:str, # The (string) name of the target variable; e.g., for Titanic, it's "Survived"
                       class_names:list=None, # List of names associated with the labels (same order); e.g. ['no','yes']
                       tree_index=0, # Index (from 0) of tree if model is an ensemble of trees like a random forest.
                       depth_range_to_display=None, # Range of depth levels to be displayed. The range values are inclusive
                       fancy=False, # To draw fancy tree chart (as opposed to simplified one)
                       scale=1.0 # Scale of the chart. Higher means bigger
                      ):
    "Plot a decision tree using dtreeviz. Note that you need to install graphviz before using this function"
    viz = dtreeviz.model(estimator,X,y,tree_index=tree_index,target_name=target_name,
                        feature_names=X.columns.values,
                        class_names=class_names,
                        )
    
    return viz.view(depth_range_to_display=depth_range_to_display,
         orientation='LR',
         instance_orientation='LR',fancy=fancy,scale=scale)

# %% ../nbs/03_chart_plotting.ipynb 57
def plot_classification_tree_sklearn(estimator, # sklearn's classifier
                      feature_names, # List of names of dependent variables (features)
                      class_names:list, # List of names associated with the labels (same order); e.g. ['no','yes']
                      rotate=True, # To rotate the tree graph
                      fname='tmp' # Name of the png file to save(no extension)
                     ):
    "Plot a decision tree classifier using sklearn. Note that this will output a png file with fname instead of showing it in the notebook"
    s = export_graphviz(estimator,out_file=None,feature_names=feature_names,filled=True,class_names=class_names,
                       special_characters=True,rotate=rotate,rounded=True)
    graph = graphviz.Source(s,format='png')
    
    _path = Path('./images')
    create_dir(_path)
    graph.render(_path/fname)

# %% ../nbs/03_chart_plotting.ipynb 65
def plot_feature_importances(importances, # feature importances from sklearn's **feature_importances_** variable
                             feature_names, # List of names of dependent variables (features)
                             figsize=(20,10), # Matplotlib figsize
                             top_n=None # Show top n features
                            ):
    "Plot and return a dataframe of feature importances, using sklearn's **feature_importances_** value"
    fea_imp_df = pd.DataFrame(data={'Feature':feature_names,'Importance':importances}).set_index('Feature')
    fea_imp_df = fea_imp_df.sort_values('Importance', ascending=True)
    if top_n is not None:
        fea_imp_df = fea_imp_df.tail(top_n)
    fea_imp_df.plot(kind='barh',figsize=figsize)
    plt.show()
    return fea_imp_df

# %% ../nbs/03_chart_plotting.ipynb 69
def plot_permutation_importances(model, # sklearn tree model that has been trained
                                 X, # Training features
                                 y, # Training label
                                 scoring=['f1_macro'], # metric, or a list of metrics
                                 n_repeats=5, # Number of times to permute a feature (higher means more accurate)
                                 seed=42, # Random seed
                                 top_n=None, # Show top n features
                                 figsize=(20,10) # Matplotlib figsize
                                ):
    "Plot and return a dataframe of feature importances using sklearn.inspection.permutation_importance"
    scoring = val2list(scoring)
    r_multi = permutation_importance(model, X, y, n_repeats=n_repeats, random_state=seed, scoring=scoring)
    fea_imp_dfs=[]
    for metric in r_multi.keys():
        print(f"{metric}")
        r = r_multi[metric]        
        fea_imp_df  = pd.DataFrame(data={'Feature':X.columns.values,'Importance':r['importances_mean'],'STD':r['importances_std']}).set_index('Feature')
        fea_imp_df = fea_imp_df.sort_values(['Importance'],ascending=True)
        if top_n is not None:
            fea_imp_df = fea_imp_df.tail(top_n)
        fig, ax = plt.subplots()
        fea_imp_df['Importance'].plot(kind='barh',figsize=figsize,xerr=fea_imp_df['STD'],ax=ax)
        ax.set_title("Feature importances using permutation")
        ax.set_ylabel("Mean metric decrease")
        fig.tight_layout()
        plt.show()
        fea_imp_dfs.append(fea_imp_df)
    return fea_imp_dfs

# %% ../nbs/03_chart_plotting.ipynb 73
def params_2D_heatmap(search_cv:dict, # A dict with keys as column headers and values as columns. Typically an attribute (**cv_results_**) of GridSearchCV or RandomizedSearchCV
                      param1:str, # Name of the first hyperparameter
                      param2:str, # Name of the second hyperparameter
                      scoring:str='f1_macro', # Metric name 
                      log_param1=False, # To log the first hyperparameter
                      log_param2=False, # To log the second hyperparameter
                      figsize=(20,10), # Matplotlib figsize
                      min_hm=None, # Minimum value for the metric to show
                      max_hm=None, # Maximum value of the metric to show
                      higher_is_better=True): # Set if high metric is better
    "Plot 2D graph of metric value for each pair of hyperparameters"
    rs_df = pd.DataFrame(search_cv)
    cm = plt.cm.get_cmap('RdYlBu')
    z = rs_df[f'mean_test_{scoring}'].values
    if not higher_is_better:
        z*=-1
    x = rs_df[f'param_{param1}'].values
    y = rs_df[f'param_{param2}'].values
    plt.figure(figsize=figsize)
    sc = plt.scatter(x, y, c=z, vmin=z.min() if not min_hm else min_hm, vmax=z.max() if not max_hm else max_hm, s=20, cmap=cm)
    if log_param2:
        plt.yscale('log')
    if log_param1:
        plt.xscale('log')
    plt.colorbar(sc)
    plt.xlabel(param1)
    plt.ylabel(param2)
    plt.grid()
    plt.show()

# %% ../nbs/03_chart_plotting.ipynb 78
def params_3D_heatmap(search_cv:dict, # A dict with keys as column headers and values as columns. Typically an attribute (**cv_results_**) of GridSearchCV or RandomizedSearchCV
                      param1:str, # Name of the first hyperparameter
                      param2:str, # Name of the second hyperparameter
                      param3:str, # Name of the third hyperparameter
                      scoring:str='f1_macro', # Metric name 
                      log_param1=False, # To log the first hyperparameter
                      log_param2=False, # To log the second hyperparameter
                      log_param3=False # To log the third hyperparameter
                     ):
    "Plot 3D graph of metric value for each triplet of hyperparameters"
    rs_df = pd.DataFrame(search_cv)
    scores = rs_df[f'mean_test_{scoring}'].values
    fig = px.scatter_3d(rs_df, x=f'param_{param1}', y=f'param_{param2}', z=f'param_{param3}',
                    color=f'mean_test_{scoring}',range_color=[scores.min(),scores.max()],
                       log_x = log_param1, log_y=log_param2, log_z=log_param3)
    fig.show()

# %% ../nbs/03_chart_plotting.ipynb 82
def pdp_numerical_only(model, # sklearn tree model that has been trained
                       X:pd.DataFrame, # dataframe to perform pdp
                       num_features:list, # A list of numerical features
                       class_names:list, # List of names associated with the labels (same order); e.g. ['no','yes']
                       y_colors=None, # List of colors associated with class_names
                       ncols=2,
                       nrows=2,
                       figsize=(20,16)):
    "Plot PDP plot for numerical dependent variables"
    common_params = {
    "subsample": 40,
    "n_jobs": -1,
    "grid_resolution": 100,
    "random_state": 42,
    }
    
    num_features = val2list(num_features)
    features_info = {
        # features of interest
        "features": num_features,
        # type of partial dependence plot
        "kind": "average",
        # information regarding categorical features
        "categorical_features": None,
    }
    
    if ncols*nrows!=len(num_features):
        raise ValueError('Make sure the product of ncols and nrows is exactly the number of numerical features you want to plot')
    _, ax = plt.subplots(ncols=ncols, nrows=nrows, squeeze=False,figsize=figsize,constrained_layout=True)
    
    if y_colors is None:
        y_colors = matplotlib.colormaps['tab10'].colors

    for i in range(len(class_names)):
        _display = PartialDependenceDisplay.from_estimator(
                            model,
                            X,
                            **features_info,
                            target=i,
                            line_kw={"label": class_names[i],'color':y_colors[i]},
                            ax=ax,
                            **common_params
        )

# %% ../nbs/03_chart_plotting.ipynb 89
def pdp_categorical_only(model, # sklearn tree model that has been trained
                         X:pd.DataFrame, # dataframe to perform pdp
                         cat_feature:list, # A single categorical feature
                         class_names:list, # List of names associated with the labels (same order); e.g. ['no','yes']
                         y_colors=None, # List of colors associated with class_names
                         ymax=0.5,
                         figsize=(20,8)):
    "Plot PDP plot for categorical dependent variables"
    common_params = {
    "subsample": 40,
    "n_jobs": 2,
    "grid_resolution": 100,
    "random_state": 42,
    }
    
    features_info = {
        # features of interest
        "features": [cat_feature],
        # type of partial dependence plot
        "kind": "average",
        # information regarding categorical features
        "categorical_features": [cat_feature],
    }
    if y_colors is None:
        y_colors = matplotlib.colormaps['tab10'].colors

    displays=[]
    for i in range(len(class_names)): 
        _, ax = plt.subplots(figsize=(1,1))

        _display = PartialDependenceDisplay.from_estimator(
            model,
            X,
            **features_info,
            target=i,
            line_kw={"label": class_names[i],'color':y_colors[i]},
            ax=ax,
            **common_params,
        );
        displays.append(_display)
    
    fig, axes = plt.subplots(1, len(displays), figsize=figsize,sharey=True)
    for i in range(len(displays)):
        displays[i].plot(ax=axes[i],pdp_lim={1: (0, ymax)})
        axes[i].set_title(class_names[i])

# %% ../nbs/03_chart_plotting.ipynb 92
def plot_ice_pair(model, # sklearn tree model that has been trained
                  X:pd.DataFrame, # dataframe to perform ice
                  pair_features:list, # a list of only 2 features
                  class_idx, # index of the class to plot
                  figsize=(10,4)):
    "Plot ICE plot from a pair of numerical feature"
    common_params = {
        "subsample": 40,
        "n_jobs": -1,
        "grid_resolution": 100,
        "random_state": 42,
    }
    fea_to_plot = pair_features
    features_info = {
        # features of interest
        "features": [fea_to_plot[0],fea_to_plot[1], fea_to_plot],
        # type of partial dependence plot
        "kind": "average",
        # information regarding categorical features
        "categorical_features": None,
    }

    _, ax = plt.subplots(ncols=3, figsize=figsize, constrained_layout=True)

    _display = PartialDependenceDisplay.from_estimator(
        model,
        X,
        **features_info,
        target=class_idx,
        ax=ax,
        **common_params,
    )
    plt.setp(_display.deciles_vlines_, visible=False)

# %% ../nbs/03_chart_plotting.ipynb 98
def plot_confusion_matrix(y_true:list|np.ndarray, # A list/numpy array of true labels 
                          y_pred:list|np.ndarray, # A list/numpy array of predictions
                          labels=None # Display names matching the labels (same order).
                         ):
    "Simple function to plot the confusion matrix"
    cm = confusion_matrix(y_true, y_pred)
    disp = ConfusionMatrixDisplay(confusion_matrix=cm,
                                  display_labels=labels)
    disp.plot()
    plt.show()

# %% ../nbs/03_chart_plotting.ipynb 102
def draw_sankey(data, target,chart_name,save_name=None):
    PATH = Path('sk_reports')
    unique_source_target = list(pd.unique(data[['source', 'target']].values.ravel('K')))
    mapping_dict = {k: v for v, k in enumerate(unique_source_target)}
    data = data.copy()
    data['source'] = data['source'].map(mapping_dict)
    data['target'] = data['target'].map(mapping_dict)
    links_dict = data.to_dict(orient='list')

    nodes = np.unique(data[["source", "target"]], axis=None)
    nodes = pd.Series(index=nodes, data=range(len(nodes)))


    fig = go.Figure(data=[go.Sankey(
        node = dict(
          pad = 15,
          thickness = 20,
          line = dict(color = "black", width = 0.5),
          label = unique_source_target,
          # color = [px.colors.qualitative.Plotly[i % len(px.colors.qualitative.Plotly)] for i in nodes.loc[data["source"]]]
        ),
        link = dict(
          source = links_dict["source"],
          target = links_dict["target"],
          value = links_dict[target],
          color = [px.colors.qualitative.Plotly[i % len(px.colors.qualitative.Plotly)] for i in nodes.loc[data["source"]]],
      ))])
    fig.update_layout(title_text=chart_name, font_size=10)
    fig.show()
    if save_name: fig.write_html(str(PATH/save_name)+ '.html')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_data_preprocess.ipynb.

# %% ../nbs/01_data_preprocess.ipynb 4
from __future__ import annotations
from .utils import *
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
from statsmodels.tools.tools import add_constant
from statsmodels.stats.outliers_influence import variance_inflation_factor
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import MinMaxScaler,StandardScaler

# %% auto 0
__all__ = ['process_missing_values', 'scale_num_cols', 'one_hot_cat', 'preprocessing_general']

# %% ../nbs/01_data_preprocess.ipynb 5
def process_missing_values(X_train:pd.DataFrame, # Training dataframe
                           X_test:pd.DataFrame=None, # Testing dataframe
                           missing_cols:list|str=[], # A column name having missing values, or a list of such columns
                           missing_vals:list|int|float|str=np.NaN, # A placeholder for missing values, or a list of placeholders for all columns in miss_cols
                           strategies:list|str='median', # The imputation strategy from sklearn, or a list of such values. Currently support 'median','mean','most_frequent'
                           **kwargs):
    "Process columns with missing values using Sklearn SimpleInputer"
    if missing_cols==[]:
        return X_train if X_test is None else (X_train,X_test)
    missing_cols = val2list(missing_cols)
    missing_vals = val2list(missing_vals,len(missing_cols))
    strategies = val2list(strategies,len(missing_cols))
    X_train = X_train.copy()
    if X_test is not None: X_test = X_test.copy()
    for i,c in enumerate(missing_cols):
        imp = SimpleImputer(missing_values=missing_vals[i], strategy=strategies[i])
        X_train[c] = imp.fit_transform(X_train[c].values.reshape(-1,1)).flatten()
        if X_test is not None: X_test[c] = imp.transform(X_test[c].values.reshape(-1,1)).flatten()
    return X_train if X_test is None else (X_train,X_test)

# %% ../nbs/01_data_preprocess.ipynb 12
def scale_num_cols(X_train:pd.DataFrame, # Training dataframe
                   X_test:pd.DataFrame=None, # Testing dataframe
                   num_cols:list|str=[], # Name of the numerical column, or a list of such columns
                   scale_methods:list|str='minmax', # Sklearn scaling method ('minmax' or 'standard'), or a list of such methods        
                    **kwargs):
    "Scale numerical columns using Sklearn"
    if num_cols==[]:
        return X_train if X_test is None else (X_train,X_test)
    num_cols = val2list(num_cols)
    scale_methods = val2list(scale_methods,len(num_cols))
    X_train = X_train.copy()
    if X_test is not None: X_test = X_test.copy()
    for i,c in enumerate(num_cols):
        if scale_methods[i]=='minmax':
            imp = MinMaxScaler()
        elif scale_methods[i]=='standard':
            imp = StandardScaler()
        else:
            raise ValueError('Unrecognized scaling method. Accept methods: minmax and standard')
        X_train[c] = imp.fit_transform(X_train[c].values.reshape(-1,1)).flatten()
        if X_test is not None: X_test[c] = imp.transform(X_test[c].values.reshape(-1,1)).flatten()
    return X_train if X_test is None else (X_train,X_test)

# %% ../nbs/01_data_preprocess.ipynb 16
def one_hot_cat(X_train:pd.DataFrame, # Training dataframe
                X_test:pd.DataFrame=None, # Testing dataframe
                cat_cols:list|str=[], # Name of the categorical columns (non-binary), or a list of such columns
                bi_cols:list|str=[], # Name of the binary column, or a list of such columns
                **kwargs):
    "Perform 'get_dummies' on categorical columns"
    if cat_cols==[] and bi_cols==[]:
        return X_train if X_test is None else (X_train,X_test)
    cat_cols = val2list(cat_cols)
    bi_cols = val2list(bi_cols)
    n_train = X_train.shape[0]
    if X_test is not None:
        X_total = pd.concat([X_train,X_test],axis=0)
    else:
        X_total = X_train.copy()
    if len(cat_cols):
        X_total = pd.get_dummies(X_total,columns=cat_cols,drop_first=False)
    if len(bi_cols):
        X_total = pd.get_dummies(X_total,columns=bi_cols,drop_first=True)
    return X_total if X_test is None else (X_total.iloc[:n_train].copy(), X_total.iloc[n_train:].copy())

# %% ../nbs/01_data_preprocess.ipynb 20
def preprocessing_general(X_train:pd.DataFrame, # Training dataframe
                          X_test:pd.DataFrame=None, # Testing dataframe
                          **kwargs, # Keyword arguments for processing missing values, scaling numerical columns and one-hot-encoding categorical columns
                         ):
    """
    The main preprocessing functions, will perform:
    
    - Fill missing values
    
    - Scale numerical columns
    
    - One-hot encode categorical columns
    
    Remember to put in the appropriate keyword arguments for each of the preprocessings mentioned above
    """
    results = process_missing_values(X_train,X_test,**kwargs)
    results = scale_num_cols(results,**kwargs) if X_test is None else scale_num_cols(*results,**kwargs)
    results = one_hot_cat(results,**kwargs) if X_test is None else one_hot_cat(*results,**kwargs)
    return results



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_utils.ipynb.

# %% ../nbs/05_utils.ipynb 4
from __future__ import annotations
from pathlib import Path
import numpy as np

# %% auto 0
__all__ = ['val2list', 'create_dir']

# %% ../nbs/05_utils.ipynb 5
def val2list(val,lsize=1):
    "Convert an element (nonlist value) to a list of 1 element"
    if not isinstance(val,list) and not isinstance(val,np.ndarray):
        val=[val for i in range(lsize)]
    return val

# %% ../nbs/05_utils.ipynb 10
def create_dir(path_dir):
    path_dir = Path(path_dir)
    if not path_dir.exists():
        path_dir.mkdir(parents=True)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/that-ml-library',
                'doc_host': 'https://anhquan0412.github.io',
                'git_url': 'https://github.com/anhquan0412/that-ml-library',
                'lib_path': 'that_ml_library'},
  'syms': { 'that_ml_library.chart_plotting': { 'that_ml_library.chart_plotting._cramer_v': ( 'chart_plotting.html#_cramer_v',
                                                                                              'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.draw_sankey': ( 'chart_plotting.html#draw_sankey',
                                                                                                'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.get_cat_correlation': ( 'chart_plotting.html#get_cat_correlation',
                                                                                                        'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.get_correlation_by_threshold': ( 'chart_plotting.html#get_correlation_by_threshold',
                                                                                                                 'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.get_vif': ( 'chart_plotting.html#get_vif',
                                                                                            'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.params_2D_heatmap': ( 'chart_plotting.html#params_2d_heatmap',
                                                                                                      'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.params_3D_heatmap': ( 'chart_plotting.html#params_3d_heatmap',
                                                                                                      'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.pdp_categorical_only': ( 'chart_plotting.html#pdp_categorical_only',
                                                                                                         'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.pdp_numerical_only': ( 'chart_plotting.html#pdp_numerical_only',
                                                                                                       'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_cat_correlation': ( 'chart_plotting.html#plot_cat_correlation',
                                                                                                         'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_classification_tree_sklearn': ( 'chart_plotting.html#plot_classification_tree_sklearn',
                                                                                                                     'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_confusion_matrix': ( 'chart_plotting.html#plot_confusion_matrix',
                                                                                                          'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_feature_importances': ( 'chart_plotting.html#plot_feature_importances',
                                                                                                             'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_ice_pair': ( 'chart_plotting.html#plot_ice_pair',
                                                                                                  'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_learning_curve': ( 'chart_plotting.html#plot_learning_curve',
                                                                                                        'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_permutation_importances': ( 'chart_plotting.html#plot_permutation_importances',
                                                                                                                 'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_prediction_distribution': ( 'chart_plotting.html#plot_prediction_distribution',
                                                                                                                 'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_residuals': ( 'chart_plotting.html#plot_residuals',
                                                                                                   'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_tree_dtreeviz': ( 'chart_plotting.html#plot_tree_dtreeviz',
                                                                                                       'that_ml_library/chart_plotting.py'),
                                                'that_ml_library.chart_plotting.plot_validation_curve': ( 'chart_plotting.html#plot_validation_curve',
                                                                                                          'that_ml_library/chart_plotting.py')},
            'that_ml_library.data_preprocess': { 'that_ml_library.data_preprocess.one_hot_cat': ( 'data_preprocess.html#one_hot_cat',
                                                                                                  'that_ml_library/data_preprocess.py'),
                                                 'that_ml_library.data_preprocess.preprocessing_general': ( 'data_preprocess.html#preprocessing_general',
                                                                                                            'that_ml_library/data_preprocess.py'),
                                                 'that_ml_library.data_preprocess.process_missing_values': ( 'data_preprocess.html#process_missing_values',
                                                                                                             'that_ml_library/data_preprocess.py'),
                                                 'that_ml_library.data_preprocess.scale_num_cols': ( 'data_preprocess.html#scale_num_cols',
                                                                                                     'that_ml_library/data_preprocess.py')},
            'that_ml_library.ml_helpers': { 'that_ml_library.ml_helpers.do_param_search': ( 'ml_helpers.html#do_param_search',
                                                                                            'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.get_adaboost_info': ( 'ml_helpers.html#get_adaboost_info',
                                                                                              'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.run_logistic_regression': ( 'ml_helpers.html#run_logistic_regression',
                                                                                                    'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.run_multinomial_statmodel': ( 'ml_helpers.html#run_multinomial_statmodel',
                                                                                                      'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.run_sklearn_model': ( 'ml_helpers.html#run_sklearn_model',
                                                                                              'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.show_both_cv': ( 'ml_helpers.html#show_both_cv',
                                                                                         'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.summarize_cv_results': ( 'ml_helpers.html#summarize_cv_results',
                                                                                                 'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.summarize_default_cv': ( 'ml_helpers.html#summarize_default_cv',
                                                                                                 'that_ml_library/ml_helpers.py'),
                                            'that_ml_library.ml_helpers.tune_sklearn_model': ( 'ml_helpers.html#tune_sklearn_model',
                                                                                               'that_ml_library/ml_helpers.py')},
            'that_ml_library.utils': { 'that_ml_library.utils.create_dir': ('utils.html#create_dir', 'that_ml_library/utils.py'),
                                       'that_ml_library.utils.val2list': ('utils.html#val2list', 'that_ml_library/utils.py')}}}
